@@ -29,6 +29,14 @@ ResolvedSerial = 'rs'
 class BadClassName(Exception):
     pass
 
+class BadClass(object):
+
+    def __init__(self, *args):
+        self.args = args
+
+    def __reduce__(self):
+        raise BadClassName(*self.args)
+
 _class_cache = {}
 _class_cache_get = _class_cache.get
 def find_global(*args):
@@ -48,11 +56,18 @@ def find_global(*args):
 
     if cls == 1:
         # Not importable
-        raise BadClassName(*args)
+        if (isinstance(args, tuple) and len(args) == 2 and
+            isinstance(args[0], basestring) and
+            isinstance(args[1], basestring)
+            ):
+            return BadClass(*args)
+        else:
+            raise BadClassName(*args)
     return cls
 
 def state(self, oid, serial, prfactory, p=''):
     p = p or self.loadSerial(oid, serial)
+    p = self._crs_untransform_record_data(p)
     file = StringIO(p)
     unpickler = Unpickler(file)
     unpickler.find_global = find_global
@@ -80,13 +95,13 @@ class IPersistentReference(zope.interface.Interface):
 
     def __cmp__(other):
         '''if other is equivalent reference, return 0; else raise ValueError.
-        
+
         Equivalent in this case means that oid and database_name are the same.
 
         If either is a weak reference, we only support `is` equivalence, and
         otherwise raise a ValueError even if the datbase_names and oids are
         the same, rather than guess at the correct semantics.
-        
+
         It is impossible to sort reliably, since the actual persistent
         class may have its own comparison, and we have no idea what it is.
         We assert that it is reasonably safe to assume that an object is
@@ -108,7 +123,13 @@ class PersistentReference(object):
         self.data = data
         # see serialize.py, ObjectReader._persistent_load
         if isinstance(data, tuple):
-            self.oid, self.klass = data
+            self.oid, klass = data
+            if isinstance(klass, BadClass):
+                # We can't use the BadClass directly because, if
+                # resolution succeeds, there's no good way to pickle
+                # it.  Fortunately, a class reference in a persistent
+                # reference is allowed to be a module+name tuple.
+                self.data = self.oid, klass.args
         elif isinstance(data, str):
             self.oid = data
         else: # a list
@@ -116,13 +137,17 @@ class PersistentReference(object):
             # 'm' = multi_persistent: (database_name, oid, klass)
             # 'n' = multi_oid: (database_name, oid)
             # 'w' = persistent weakref: (oid)
+            #    or persistent weakref: (oid, database_name)
             # else it is a weakref: reference_type
             if reference_type == 'm':
-                self.database_name, self.oid, self.klass = data[1]
+                self.database_name, self.oid, _ = data[1]
             elif reference_type == 'n':
                 self.database_name, self.oid = data[1]
             elif reference_type == 'w':
-                self.oid, = data[1]
+                try:
+                    self.oid, = data[1]
+                except ValueError:
+                    self.oid, self.database_name = data[1]
                 self.weak = True
             else:
                 assert len(data) == 1, 'unknown reference format'
@@ -131,7 +156,7 @@ class PersistentReference(object):
 
     def __cmp__(self, other):
         if self is other or (
-            isinstance(other, PersistentReference) and 
+            isinstance(other, PersistentReference) and
             self.oid == other.oid and
             self.database_name == other.database_name and
             not self.weak and
@@ -148,6 +173,16 @@ class PersistentReference(object):
     def __getstate__(self):
         raise PicklingError("Can't pickle PersistentReference")
 
+
+    @property
+    def klass(self):
+        # for tests
+        data = self.data
+        if isinstance(data, tuple):
+            return data[1]
+        elif isinstance(data, list) and data[0] == 'm':
+            return data[1][2]
+
 class PersistentReferenceFactory:
 
     data = None
@@ -175,6 +210,7 @@ def tryToResolveConflict(self, oid, committedSerial, oldSerial, newpickle,
     # class_tuple, old, committed, newstate = ('',''), 0, 0, 0
     try:
         prfactory = PersistentReferenceFactory()
+        newpickle = self._crs_untransform_record_data(newpickle)
         file = StringIO(newpickle)
         unpickler = Unpickler(file)
         unpickler.find_global = find_global
@@ -190,30 +226,42 @@ def tryToResolveConflict(self, oid, committedSerial, oldSerial, newpickle,
             newargs = ()
 
         if klass in _unresolvable:
-            return None
+            raise ConflictError
 
-        newstate = unpickler.load()
         inst = klass.__new__(klass, *newargs)
 
         try:
             resolve = inst._p_resolveConflict
         except AttributeError:
             _unresolvable[klass] = 1
-            return None
+            raise ConflictError
 
-        old = state(self, oid, oldSerial, prfactory)
+
+        oldData = self.loadSerial(oid, oldSerial)
+        if not committedData:
+            committedData  = self.loadSerial(oid, committedSerial)
+
+        if newpickle == oldData:
+            # old -> new diff is empty, so merge is trivial
+            return committedData
+        if committedData == oldData:
+            # old -> committed diff is empty, so merge is trivial
+            return newpickle
+
+        newstate = unpickler.load()
+        old       = state(self, oid, oldSerial, prfactory, oldData)
         committed = state(self, oid, committedSerial, prfactory, committedData)
 
         resolved = resolve(old, committed, newstate)
 
         file = StringIO()
         pickler = Pickler(file,1)
-        pickler.persistent_id = persistent_id
+        pickler.inst_persistent_id = persistent_id
         pickler.dump(meta)
         pickler.dump(resolved)
-        return file.getvalue(1)
+        return self._crs_transform_record_data(file.getvalue(1))
     except (ConflictError, BadClassName):
-        return None
+        pass
     except:
         # If anything else went wrong, catch it here and avoid passing an
         # arbitrary exception back to the client.  The error here will mask
@@ -221,9 +269,19 @@ def tryToResolveConflict(self, oid, committedSerial, oldSerial, newpickle,
         # ConflictError, but not necessarily from other errors.  But log
         # the error so that any problems can be fixed.
         logger.error("Unexpected error", exc_info=True)
-        return None
 
-class ConflictResolvingStorage:
+    raise ConflictError(oid=oid, serials=(committedSerial, oldSerial),
+                        data=newpickle)
+
+class ConflictResolvingStorage(object):
     "Mix-in class that provides conflict resolution handling for storages"
 
     tryToResolveConflict = tryToResolveConflict
+
+    _crs_transform_record_data = _crs_untransform_record_data = (
+        lambda self, o: o)
+
+    def registerDB(self, wrapper):
+        self._crs_untransform_record_data = wrapper.untransform_record_data
+        self._crs_transform_record_data = wrapper.transform_record_data
+        super(ConflictResolvingStorage, self).registerDB(wrapper)
